#include <cppunit/extensions/HelperMacros.h>

#include <arc/notify.h>

#include <sstream>

class NotifyTest : public CppUnit::TestFixture {

	CPPUNIT_TEST_SUITE( NotifyTest );
	CPPUNIT_TEST( TestNotifierINFO );
	CPPUNIT_TEST( TestNotifierDEBUG );
	CPPUNIT_TEST_SUITE_END();

	public:
		void setUp();
		void tearDown();

		void TestNotifierINFO();
		void TestNotifierDEBUG();

	private:
		std::stringstream* output;
};


void NotifyTest::setUp() {
	output = new std::stringstream;
	Notify::getNotifier()->SetOutStream(*output);
}


void NotifyTest::tearDown() {
	delete output;
}


void NotifyTest::TestNotifierINFO() {
	notify(DEBUG) << "This DEBUG message should not be seen" << std::endl;
	CPPUNIT_ASSERT_EQUAL( (int)output->str().size(), 0 );

	notify(INFO) << "This INFO message should be seen" << std::endl;
	CPPUNIT_ASSERT_EQUAL( output->str(), std::string("This INFO message should be seen\n") );
	output->str("");
}


void NotifyTest::TestNotifierDEBUG() {
	SetNotifyLevel(DEBUG);
	notify(DEBUG) << "This DEBUG message should now be seen" << std::endl;
	CPPUNIT_ASSERT_EQUAL( output->str(), std::string("This DEBUG message should now be seen\n") );
	output->str("");

	notify(INFO) << "This INFO message should also be seen" << std::endl;
	CPPUNIT_ASSERT_EQUAL( output->str(), std::string("This INFO message should also be seen\n") );
	output->str("");
}


CPPUNIT_TEST_SUITE_REGISTRATION( NotifyTest );
