#include <iostream>
#include <list>
#include <string>
#include <algorithm>

#include "arccli.h"

#include <arc/ftpcontrol.h>
#include <arc/jobftpcontrol.h>
#include <arc/joblist.h>
#include <arc/mdsquery.h>
#include <arc/notify.h>
#include <arc/resource.h>

#ifdef HAVE_LIBINTL_H
#include <libintl.h>
#define _(A) dgettext("arccli", (A))
#else
#define _(A) (A)
#endif


void arcget(const std::list<std::string>& jobs,
            const std::list<std::string>& clusterselect,
            const std::list<std::string>& clusterreject,
            const std::list<std::string>& status,
            const std::string& directory,
            const bool usejobname,
            const bool keep,
            const int timeout,
            const bool anonymous) {

	std::list<std::string> jobids =
		GetJobIDsList(jobs, clusterselect, clusterreject);

	if (jobs.empty() && clusterselect.empty() && jobids.empty())
		throw ARCCLIError(_("No jobs"));

	if (jobids.empty())
		throw ARCCLIError(_("No valid jobnames/jobids given"));

	std::list<Job> joblist =
		GetJobInfo(jobids, MDS_FILTER_JOBINFO, anonymous, "", timeout);

	FTPControl ctrl;
	JobFTPControl jctrl;

	for (std::list<Job>::iterator jli = joblist.begin();
	     jli != joblist.end(); jli++) {

		if (jli->status.empty()) {
			notify(WARNING) << _("Job information not found")
			                << ": " << jli->id << std::endl;
			continue;
		}

		if (!status.empty() &&
		    std::find(status.begin(), status.end(), jli->status) == status.end())
			continue;

		if (jli->status == "DELETED") {
			notify(WARNING) << _("Job has already been deleted")
			                << ": " << jli->id << std::endl;
			continue;
		}

		if (jli->status != "FINISHED" && jli->status != "FAILED" &&
		    jli->status != "KILLED") {
			notify(WARNING) << _("Job has not finished yet")
			                << ": " << jli->id << std::endl;
			continue;
		}

		std::string dir;
		if (usejobname)
			dir = directory + '/' + jli->job_name;
		else
			dir = directory + '/' + jli->id.substr(jli->id.rfind('/') + 1);

		try {
			MkDir(dir, true);
		}
		catch (ARCLibError& e) {
			notify(WARNING) << e.what() << std::endl;
			notify(WARNING) << _("Skipping job")
			                << ": " << jli->id << std::endl;
			continue;
		}

		try {
			ctrl.DownloadDirectory(jli->id, dir, timeout);

			if (!keep) {
				notify(WARNING) << _("Deleting job")
				                << ": " << jli->id << std::endl;

				jctrl.Clean(jli->id, timeout);
				RemoveJobID(jli->id);
			}
		} catch (ARCLibError& e) {
			notify(WARNING) << e.what() << std::endl;
		}
	}
}
