/*
 * Created on Apr 19, 2005
 *
 *
 */
package se.umu.wsag.impl;

import java.util.Calendar;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.Duration;
import org.apache.log4j.Logger;
import org.ggf.ws_agreement.AgreementType;
import org.ggf.ws_agreement.GuaranteeTermType;
import org.ggf.ws_agreement.ServiceDescriptionTermType;
import org.ggf.ws_agreement.TermCompositorType;
import org.ggf.ws_agreement.TermTreeType;
import org.globus.delegation.DelegationUtil;
import org.globus.gsi.GlobusCredential;
import org.globus.security.gridmap.GridMap;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.security.descriptor.ServiceSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ServiceSecurityDescriptor;
import org.globus.wsrf.security.SecurityManager;

import se.umu.reservation.CommittedType;
import se.umu.reservation.Config;
import se.umu.reservation.DecisionMakerConfig;
import se.umu.reservation.DurationType;
import se.umu.reservation.EarliestStartType;
import se.umu.reservation.ExactStartType;
import se.umu.reservation.FlexibleType;
import se.umu.reservation.LatestStartType;
import se.umu.reservation.LocalIDType;
import se.umu.reservation.LocalReservation;
import se.umu.reservation.NumberOfCPUsType;
import se.umu.reservation.ReservationDescriptionType;
import se.umu.reservation.ReservationServiceLevelObjective;
import se.umu.reservation.ReservationType;
import se.umu.reservation.StringMap;
import se.umu.wsag.util.Executor;


/**
 * @author tordsson
 *
 * The ReservationDecisionMaker performs admission control
 * against the local (maui) scheduler. It is used by the 
 * AgreementFactoryService before this service creates a
 * new Agreement. The ReservationDecisionMaker is also 
 * used to remove local reservations.  
 * Its configuration is determined by the configuration file
 * for the AgreementFactoryService. 
 */
public class ReservationDecisionMaker implements AgreementDecisionMaker {
			
    private static Logger logger = 
        Logger.getLogger(ReservationDecisionMaker.class);
    
    private String createCommand; 
    private String releaseCommand; 
    private int initialAgreementLifeTime; 
    
    private LocalAgreementMapping agreementMapping; 
  
    // Sets up properties specific for the implementation of 
    // DecisionMaker, in this case, the commands used to 
    // create and release reservations.   
    // Also, get the class used to create the local mapping, 
    // in the ARC case, associating the proxy used to create 
    // the reservation with the reservation ID. 
    public void init(Object context) throws Exception {
    	
    	DecisionMakerConfig decisionConfig = (DecisionMakerConfig) context; 
     	Config config = decisionConfig.getConfig(); 
    	
    	createCommand = null; 
    	releaseCommand = null;     	    	
    	
    	this.initialAgreementLifeTime = decisionConfig.
				getInitialAgreementLifeTime(); 
    	if(this.initialAgreementLifeTime == 0)
    		this.initialAgreementLifeTime = 60; 
    	
    	logger.debug("Initial life time is: " + this.initialAgreementLifeTime); 
    	
    	StringMap params[] = config.getContext(); 
    
    	String localMappingClassName = null; 
    	String mappingContext = null; 
    	for(int i = 0; i < params.length; i++) {
    		String name = params[i].getName(); 
    		String value = params[i].getValue(); 
    		if("createCommand".equalsIgnoreCase(name))
    			createCommand = value; 
    		else if("releaseCommand".equalsIgnoreCase(name))
    			releaseCommand = value;
    		else if("localMappingClass".equals(name))
    			localMappingClassName = value; 
    		else if("mappingContext".equalsIgnoreCase(name))
    			mappingContext = value; 
    		else {
    			logger.warn("Unknown configuration parameter \"" + name + "\""); 
    		}
    	}
    	
    	if(releaseCommand == null || createCommand == null) 
    		throw new Exception("Invalid configuration, must contain both "
    				+ "\"requestCommand\" and  \"releaseCommand\"");     	    	
    	if(localMappingClassName == null || mappingContext == null) 
    		throw new Exception("Invalid configuration, must contain both "
    				+ "\"localMappingClass\" and \"mappingContext\""); 
    	
    	// Set up the localAgreementMapping
    	this.agreementMapping = (LocalAgreementMapping) Class.forName(
    			localMappingClassName).newInstance(); 
    	this.agreementMapping.init(mappingContext); 
    	
    	logger.debug("Local mapping class is: " +this.agreementMapping.getClass()); 
    	
    }
    
    /**
     * @return The configuration of this decision maker, 
     * useful for saving the decision maker on persistant storage. 
     */
    public DecisionMakerConfig getConfig() {
    	
    	DecisionMakerConfig decisionConfig = new DecisionMakerConfig(); 
    	decisionConfig.setClassName(ReservationDecisionMaker.class.getName()); 
    	decisionConfig.setInitialAgreementLifeTime(
    			this.initialAgreementLifeTime); 
    	
    	StringMap createContext = new StringMap(); 
    	createContext.setName("createCommand"); 
    	createContext.setValue(createCommand); 
    	StringMap releaseContext = new StringMap(); 
    	releaseContext.setName("releaseCommand"); 
    	releaseContext.setValue(releaseCommand);     	
    	StringMap context [] = new StringMap[] {createContext, releaseContext}; 
    	
    	Config config = new Config(); 
    	config.setContext(context); 
    	
    	decisionConfig.setConfig(config);
    	
    	return decisionConfig; 
    }
	
    /**
     * Extracts reservation-specific parameters from the offer. 
     * @param offer
     * @return
     */
	public ReservationType extractRequest(AgreementType offer) {
		try {
		
			ReservationType reservationOffer = new ReservationType(); 
			
			TermTreeType terms = offer.getTerms(); 
			TermCompositorType all = terms.getAll();
			
			// ServiceDescriptionTerms: 			
			ServiceDescriptionTermType term = all.getServiceDescriptionTerm();				
			
			MessageElement any[] = term.get_any(); 			
			ReservationDescriptionType reservationDescription = 
				(ReservationDescriptionType) ObjectDeserializer.
				toObject(any[0], ReservationDescriptionType.class);
			
			//NumberOfCPUsType noCPUs = (NumberOfCPUsType) ObjectDeserializer.
			//	toObject(any[0], NumberOfCPUsType.class);
			NumberOfCPUsType noCPUs = reservationDescription.getNumberOfCPUs(); 			
			reservationOffer.setNumberOfCPUs(noCPUs);
			logger.debug("Requested: " + noCPUs.getNumberOfCPUs().toString()
					+ " CPUs"); 		
			
			FlexibleType flexible = reservationDescription.getFlexible(); 
			reservationOffer.setFlexible(flexible); 
			logger.debug("Reservation is " + (flexible.isFlexible()? "" : "not") 
					+ " flexible"); 
			
			DurationType duration = reservationDescription.getDuration(); 
			reservationOffer.setDuration(duration); 
			logger.debug("Duration: " + duration.getDuration().getSeconds());
			
			// GuaranteeTerms (we skip service properties for now, 
			// these will come in handy when (if) XQuery (or XPath) is used
			// for interpreting offer. 
			
			GuaranteeTermType guaranteeTerms = all.getGuaranteeTerm();
			
			// Get earliest and latest start from reservationSLOs. 
			Object SLOs = guaranteeTerms.getServiceLevelObjective();
			ReservationServiceLevelObjective reservationSLOs; 
			if(SLOs instanceof ReservationServiceLevelObjective) {
				
				reservationSLOs = (ReservationServiceLevelObjective) SLOs; 
				String SLOTerms[] = reservationSLOs.getTerm();
				String SLOTerm = null; 
				
				// FIXME: This parsing is really ugly...								
				for(int i = 0; i < SLOTerms.length; i++) {					
					SLOTerm = SLOTerms[i]; 
					int pos1 = SLOTerm.indexOf(" "); 
					String termName = SLOTerm.substring(0, pos1); 
					int pos2 = SLOTerm.indexOf(" ", pos1 + 1); 
					String termOp = SLOTerm.substring(pos1 + 1, pos2); 
					String termValue = SLOTerm.substring(pos2 + 1, 
							SLOTerm.length());				
					
					if("IS_AFTER".equals(termOp) && 
							"earlyStart".equals(termName)) {
						try {
							Calendar early = Calendar.getInstance();
							long time = Long.parseLong(termValue);						
							early.setTimeInMillis(time*1000); 										
							reservationOffer.setEarliestStart(
									new EarliestStartType(early));
						} catch (Exception e) {
							reservationOffer.setEarliestStart(null); 
						}						
					}
					if("IS_BEFORE".equals(termOp) &&
							"lateStart".equals(termName)) {
						try {
							Calendar late = Calendar.getInstance(); 
							long time = Long.parseLong(termValue); 
							late.setTimeInMillis(time*1000); 
							reservationOffer.setLatestStart(
									new LatestStartType(late));
						} catch (Exception e) {
							reservationOffer.setLatestStart(null); 
						}																		
					}	
				}
			}
			else {
				logger.debug("Unknown type for ServiceLevelObjectives!");
				return null; 
			}
			
			return reservationOffer;
		} catch (Exception e) {
			return null; 
		}
			
	}
	
	/**
	 * Retrieves the local user ID for the user executing the service
	 * from the gridmap file. 
	 * @return the local user id
	 */
	public String getLocalUserID() {
		
		String localUser; 
		try {
			SecurityManager manager = SecurityManager.getManager(); 
			String globusID = manager.getCaller(); 
			
			// Get location of grid-mapfile
			ServiceSecurityDescriptor desc = ServiceSecurityConfig.
                            getSecurityDescriptor();        	
			String gridMapFileName = desc.getGridMapFile();
			
			// Get local userid
			GridMap gridMap = new GridMap(); 
			gridMap.load(gridMapFileName);         	        	
			localUser = gridMap.getUserID(globusID); 
			
			return localUser; 
			
		} catch (Exception e) {
			logger.error("Could not retrieve local id:");
			logger.error("[" + e.getMessage() + "]");         			
			return null; 
		}
	}
	
	/**
	 * This method interfaces the Maui scheduler, and checks if a reservation
	 * can be created
	 */
	public LocalReservation createLocalReservation(int numberOfCPUs, 
			Calendar earliestStart, 
			Calendar latestStart, 
			Duration duration, 
			boolean flexible, 
			GlobusCredential delegatedCredential) throws Exception {			
		
		String command = this.createCommand; 
		
	    if(numberOfCPUs <= 0) {
	    	logger.debug("Number of CPUs must be positive"); 
            throw new Exception("Number of CPUs must be positive"); 
        }
	    
	    if(duration.isNegative()) {
	    	logger.debug("Duration must be positive"); 	    
	    	throw new Exception("Duration must be positive");
	    }
	    
	    // Odd, Duration.setSeconds(int), but double Duration.getSeconds()...
	    Double d = new Double(duration.getSeconds()); 
	    int durationSeconds = d.intValue();

        command += " -n " + numberOfCPUs; 
        command += " -d " + durationSeconds; 

        long earliestStartSeconds = earliestStart.getTimeInMillis() / 1000; 
        long latestStartSeconds = latestStart.getTimeInMillis() / 1000; 

        // Earliest and latest job start. 
        command += " -e " + earliestStartSeconds;
        command += " -l " + latestStartSeconds; 
        
        logger.debug("Executing: " + command); 
            
        // Get local userID from gridmap 
        String localUser = null;
        localUser = getLocalUserID(); 
        if(localUser == null) {        	        
        	String gridUser  = SecurityManager.getManager().getCaller(); 
        	logger.debug("Can not find local userID for " + gridUser);
        	throw new Exception("Can not find local userID " + gridUser); 
        }             
        
        // Owner of local reservation == Caller          
        command += " -u " + localUser;  
                
        // Must be placed last
        if(flexible) 
            command += " -f"; 

        // Run command
        String output = null; 
        try {
            output = Executor.runCommand(command); 
        } catch (Exception e) {
            logger.error("Error executing " + command); 
            throw e;
        }

        // Parse output, format is either "failed" or
        // reservationid starttime
        output.trim(); 
        if(output.equalsIgnoreCase("failed")) {
        	logger.debug("Local reservation could not be granted..."); 
            throw new Exception("Failed to created reservation...");
        }

        int pos = output.indexOf(" "); 
        String localID = output.substring(0, pos); //.trim(); 
        String startTimeString = output.substring(pos+1).trim(); 
        logger.info("Reservation: [" + localID + "] successfully created");
        logger.debug("Exact start time is: " + startTimeString); 
        
        Date startTimeDate = new Date(Long.parseLong(startTimeString)*1000); 

        logger.debug("Exact start as date: " + startTimeDate); 
        Calendar startTime = Calendar.getInstance(); 
        startTime.setTime(startTimeDate);
        
        logger.debug("Exact start as Cal: " + startTime.getTime()); 
        
        List l = new LinkedList(); 
        l.add(delegatedCredential); 
        l.add(localID); 
        this.agreementMapping.create(l); 
        // l.(delegatedCredential); 
        /*
        // Store user proxy for later comparision. 
        // TODO: Move this to a plugin. 
        String proxyPath = "/var/spool/nordugrid/jobstatus/" 
        	+ "reservation." + ""+ localID + ".proxy"; 
        OutputStream proxyLocation = new FileOutputStream(proxyPath);         
        logger.debug("Saving proxy for user: " + delegatedCredential.
        		getIdentity() + " to file: " + proxyPath); 
        delegatedCredential.save(proxyLocation);        
        */
        
        // Return info about local reservation
        LocalReservation reservation = new LocalReservation(); 
        reservation.setLocalID(new LocalIDType(localID));
        reservation.setExactStart(new ExactStartType(startTime)); 
        reservation.setCommitted(new CommittedType(false)); 
        
        return reservation; 		
	}
	
/**
 * Extract reservation specific terms from offer, call local scheduler, 
 * to create local reservation, and return decision. 
 */
	public AgreementDecision create(AgreementType offer) {
		
		ReservationType reservationOffer = extractRequest(offer);
 	
		if(reservationOffer == null) 
			return new AgreementDecision(false, "Could not read offer", null); 
		
		// Hmm, seems to be one level of nestling too much in these stubs...
		int numberOfCPUs = reservationOffer.getNumberOfCPUs().getNumberOfCPUs().
			intValue();		
			
		Calendar early = null; 
		Calendar late = null; 
		try {		
			early = reservationOffer.getEarliestStart().getEarliestStart();
			late = reservationOffer.getLatestStart().getLatestStart();
		} catch (Exception e) {
			logger.error("Could not determine preferred start time"); 
			early = Calendar.getInstance(); 
			late = Calendar.getInstance(); 
			early.add(Calendar.HOUR, 2);
			late.add(Calendar.HOUR, 12); 			
		}
		Duration duration = reservationOffer.getDuration().getDuration();
		boolean flexible = reservationOffer.getFlexible().isFlexible(); 
		
		// Get epr to delegated credentials, 
		// it is passed as the agreementinitiator. 
		logger.debug("Getting delegated credentials..."); 
		GlobusCredential delegatedCredential = null; 
		try {
			EndpointReferenceType delegatedCredentialEpr = 
				(EndpointReferenceType) offer.getContext().
				getAgreementInitiator(); 
			delegatedCredential = DelegationUtil.
				getDelegationResource(delegatedCredentialEpr).getCredential(); 
			logger.debug("Retrieving delegated credentials: " + 
					delegatedCredential.getIdentity()); 
		} catch (Exception e) {
			logger.error("Could not get delegated credential: " 
					+ e.getMessage()); 
			return new AgreementDecision(false, "Error creating local res" + 
					e.getMessage(), null); 		
		}
		
		logger.debug("Requesting local reservation");
		LocalReservation localReservation = null; 
		try {
			localReservation = createLocalReservation(numberOfCPUs, 
				early, 
				late, 
				duration, 
				flexible, 
				delegatedCredential);
		} catch (Exception e) {
			logger.error("Local reservation failed"); 
			return new AgreementDecision(false, "Error creating local res" + 
					e.getMessage(), null); 		
		}

	    // Pass localID and exactStart as context in AgreementDecision
		return new AgreementDecision(true, "", localReservation); 
	
	}
	
	// cancel, remove the local reservation. 
	public void cancel(ServiceDescriptionTermType terms) {
		
		String localID = null; 		
		try {
			MessageElement any[] = terms.get_any(); 
			ReservationType reservationInfo = (ReservationType) 
				ObjectDeserializer.toObject(any[0], ReservationType.class); 
			localID =  reservationInfo.getLocalID().getLocalID(); 
			logger.debug("Releasing reservation: " + localID); 
			String command = this.releaseCommand + " " + localID;
			logger.debug("Executing command [" + command + "]"); 
			Executor.runCommand(command);
			
		} catch (Exception e) {
			logger.error("Error releasing reservation " + localID); 
		}
		try {
			this.agreementMapping.cancel(localID); 
		} catch (Exception e) {
			logger.error("Error removing local proxy file for reservation: " 
					+ localID); 
		}
			/*
			// TODO: Move this code to some arc-plugin. 
			// Remove saved proxy file...
			String proxyPath = "/var/spool/nordugrid/jobstatus/" 
				+ "reservation." + localID + ".proxy"; 
			logger.debug("Removing proxy file:\n " + proxyPath); 
			File proxyFile = new File(proxyPath); 
			proxyFile.delete(); 			
			*/
			
	}

	/**
	 * @return Returns the initialAgreementLifeTime.
	 */
	public int getInitialAgreementLifeTime() {
		return initialAgreementLifeTime;
	}
}
